package ap.proof.goal;

import ap.terfor.{Term, TermOrder, ConstantTerm, Quantifier, ArithConj}
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.ColumnSolver
import ap.terfor.substitutions.ConstantSubst
import ap.util.{Logic, Debug}

case object FactsNormalisationTask extends Task {

  val AC = Debug.AC_FACTS_TASK
  
  val priority : Int = -10000 
  
  def apply(goal : Goal, taskFinishedAction : TaskFinishedAction) : ProofTree = {
    print("n")
    
    val solver = new GoalColumnSolver (goal)
    val (newEqConj, newOrder) = solver.result
    
    val newFacts = goal.facts.updatePositiveEqs(newEqConj, newOrder)
                             .internallyPropagate
    
    val newSubTree = goal.taskFinished(newFacts,
                                       solver.eliminatedConstants,
                                       newOrder,
                                       List(),
                                       taskFinishedAction)
                             
    solver.postProcessor(newSubTree)
  }
  
}


////////////////////////////////////////////////////////////////////////////////

private class GoalColumnSolver(goal : Goal)
              extends ColumnSolver(goal.facts.arithConj.positiveEqs, goal.order) {
  
  private var constCounter : Int = 0
  var eliminatedConstants = goal.eliminatedConstants
  
  /**
   * The function that is supposed to be applied to the resulting proof tree
   * (adding the <code>QuantifiedTree</code>, <code>SubstTree</code>
   * constructors)
   */
  var postProcessor : ProofTree => ProofTree = ((p) => p)
  
  //////////////////////////////////////////////////////////////////////////////
  
  protected def isSolvableEq(lc : LinearCombination) : Boolean =
    !lc.leadingCoeff.isOne && noneOrTwoElimConstants(lc)

  protected def createSmallSym(lc : LinearCombination,
                               symDefinition : LinearCombination,
                               order : TermOrder) : (Term, TermOrder) = {
    val smallConst = new ConstantTerm("sc_" + goal.age + "_" + constCounter)
    constCounter = constCounter + 1
    
    if (isEliminated(lc(0) _2)) {
      // then also the new constant can be eliminated, and has to be put in
      // between the non-eliminated and the eliminated constants
      val extendedOrder = order.extend(smallConst, eliminatedConstants)
      eliminatedConstants = eliminatedConstants + smallConst
      postProcessor = postProcessor compose
                      ((p:ProofTree) =>
                       QuantifiedTree(p, Quantifier.ALL, Set(smallConst),
                                      order, extendedOrder))
      (smallConst, extendedOrder)
    } else {
      // then it is not possible to eliminate the new constant, and the constant
      // can be made smaller than all other symbols
      val extendedOrder = order.extend(smallConst, order.orderedConstants)
      val backSubst = new ConstantSubst(smallConst, symDefinition, extendedOrder)
      postProcessor = postProcessor compose
                      ((p:ProofTree) => SubstTree(p, backSubst, order, extendedOrder))
      (smallConst, extendedOrder)
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  
  private def isEliminated(t : Term) : Boolean = t match {
    case t : ConstantTerm => eliminatedConstants contains t
    case _ => false
  }

  /**
   * Determine whether the linear combination <code>lc</code> contains either
   * none or at least two terms that are eliminated in <code>goal</code>
   */
  private def noneOrTwoElimConstants(lc : LinearCombination) : Boolean = {
    ////////////////////////////////////////////////////////////////////////////
    def post(b : Boolean) = {
      Debug.assertPost(FactsNormalisationTask.AC,
                       b ==
                       (Logic.exists(0, lc.size, (i) =>
                        Logic.exists(i+1, lc.size, (j) =>
                          isEliminated(lc(i) _2) && isEliminated(lc(j) _2))))
                        ||
                        Logic.forall(for ((_, t) <- lc.elements)
                                     yield !isEliminated(t)))
      b
    }
    ////////////////////////////////////////////////////////////////////////////

    post (if (lc.isEmpty) {
            true
          } else if (isEliminated(lc(0) _2)) {
            (lc.size >= 2) && isEliminated(lc(1) _2)
          } else {
            true
          })
  }
}
