package ap.terfor;

import scala.collection.mutable.ArrayBuffer

import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.substitutions.{IdentitySubst, ComposeSubsts,
                                VariableShiftSubst, Substitution}
import ap.util.{Debug, Logic, Seqs}

object Conjunction {
  
  val AC = Debug.AC_PROP_CONNECTIVES

  /**
   * Create a <code>Conjunction</code> from an arbitrary collection of formulas.
   * It is required that all given formulas are sorted by the specified
   * <code>order</code>.
   */
  def apply(quans : Seq[Quantifier],
            formulas : Iterator[Formula],
            order : TermOrder) : Conjunction = {
    // first split up the input formulas into the arithmetic stuff and the
    // negated parts
    val (arithConj, negatedConjs) = segregateFormulas(formulas, order)

    ////////////////////////////////////////////////////////////////////////////
    Debug.assertInt(AC, !negatedConjs.containsLiteral &&
                        !negatedConjs.containsNegatedConjunction)
    ////////////////////////////////////////////////////////////////////////////

    createHelp(quans, arithConj, negatedConjs, order)
  }

  /**
   * Take apart a list of conjuncts and create one arithmetic conjunction and
   * one conjunction of negated conjunctions
   */
  private def segregateFormulas(formulas : Iterator[Formula],
                                order : TermOrder)
                                         : (ArithConj, NegatedConjunctions) = {
    val arithFors = new ArrayBuffer[Formula]
    val negConjs = new ArrayBuffer[Conjunction]
    
    def addNegatedConjunctions(negCs : NegatedConjunctions) : unit =
      for (negC <- negCs) {
        if (negC.isLiteral) {
          val negAC = negC.arithConj
          //////////////////////////////////////////////////////////////////////
          Debug.assertInt(AC, negAC.isLiteral && negC.negatedConjs.isEmpty)
          //////////////////////////////////////////////////////////////////////
          arithFors += negAC.negate
        } else if (negC.isNegatedConjunction) {
          // remove double-negation
          addFormula(negC.negatedConjs(0))
        } else {
          negConjs += negC
        }
      }
    
    def addFormula(f : Formula) : unit =
      if (!f.isTrue) {
        f match {
        case _ : EquationConj | _ : NegEquationConj | _ : ArithConj =>
          arithFors += f          
        case c : Conjunction =>
          if (c.quans.isEmpty) {
            arithFors += c.arithConj
            addNegatedConjunctions(c.negatedConjs)
          } else {
            // if the conjunction is quantified, negate it twice and then
            // add it
            addNegatedConjunctions(NegatedConjunctions(c.negate, order))
          }
        case negCs : NegatedConjunctions => addNegatedConjunctions(negCs)
        }
      }
    
    for (f <- formulas) {
      if (f.isFalse) return (ArithConj.FALSE, NegatedConjunctions.TRUE)
      addFormula(f)
    }
    
    (ArithConj.conj(arithFors, order), NegatedConjunctions(negConjs, order))
  }
   
  def apply(quans : Seq[Quantifier],
            formulas : Iterable[Formula],
            order : TermOrder) : Conjunction =
    apply(quans, formulas.elements, order)

  /**
   * Compute a conjunction from an arbitrary set of formulas
   */
  def conj(formulas : Iterator[Formula], order : TermOrder) : Conjunction =
    apply(List(), formulas, order)    

  def conj(formulas : Iterable[Formula], order : TermOrder) : Conjunction =
    apply(List(), formulas.elements, order)  
    
  def conj(f : Formula, order : TermOrder) : Conjunction =
    apply(List(), Iterator.single(f), order)  
    
  /**
   * Compute a disjunction from an arbitrary set of formulas
   */
  def disj(formulas : Iterator[Formula], order : TermOrder) : Conjunction =
    conj(for (f <- formulas) yield conj(Iterator.single(f), order).negate,
         order).negate

  def disj(formulas : Iterable[Formula], order : TermOrder) : Conjunction =
    disj(formulas.elements, order)

  /**
   * Form the implication between two formulas
   */
  def implies(for1 : Formula, for2 : Formula, order : TermOrder) : Conjunction =
    disj(Array(negate(for1, order), for2), order)
    
  /**
   * Quantify a formula
   */
  def quantify(quans : Seq[Quantifier], f : Formula, order : TermOrder)
                                                                : Conjunction =
    apply(quans, Iterator.single(f), order)
    
  /**
   * Negate a formula
   */
  def negate(f : Formula, order : TermOrder) : Conjunction =
    apply(List(), Iterator.single(f), order).negate
    
  /**
   * Create a <code>Conjunction</code> from an arbitrary collection of formulas.
   * It is required that all given formulas are sorted by the specified
   * <code>order</code>.
   */
  def apply(quans : Seq[Quantifier],
            arithConj : ArithConj, negatedConjs : NegatedConjunctions,
            order : TermOrder) : Conjunction = {
    if (arithConj.isFalse) {
      FALSE
    } else if (negatedConjs.containsLiteral ||
               negatedConjs.containsNegatedConjunction) {
      // go back to the more general method that first splits up the formulas
      // into the different parts
      apply(quans, Array(arithConj, negatedConjs), order)
    } else {
      createHelp(quans, arithConj, negatedConjs, order)
    }
  }

  private def createHelp(quans : Seq[Quantifier],
                         arithConj : ArithConj, negatedConjs : NegatedConjunctions,
                         order : TermOrder) : Conjunction = {
    if (arithConj.isTrue && negatedConjs.isNegatedQuantifiedConjunction) {
      // pull out the quantifiers
      val negC = negatedConjs(0)
      apply((for (q <- negC.quans) yield q.dual) ++ quans,
            ArithConj.TRUE,
            NegatedConjunctions(createFromNormalised(List(),
                                                     negC.arithConj,
                                                     negC.negatedConjs,
                                                     order),
                                order),
            order)
     } else {
       createFromNormalised(quans, arithConj, negatedConjs, order)
     }
  }
   
  /**
   * Create a <code>Conjunction</code> from different parts that are already
   * normalised. This primarily means that <code>negatedConjs</code> must not
   * contain any conjunctions that are just literals, and that quantifiers are
   * pulled out completely if the conjunction only has a single conjunct.
   */
  def createFromNormalised(quans : Seq[Quantifier],
                           arithConj : ArithConj,
                           negatedConjs : NegatedConjunctions,
                           order : TermOrder) : Conjunction = {
    if (arithConj.isFalse || negatedConjs.isFalse)
      FALSE
    else if (arithConj.isTrue && negatedConjs.isTrue)
      TRUE
    else {
      val (resultingQuans, resultingSubst) =
        eliminateUnusedQuans(arithConj.variables ++ negatedConjs.variables,
                             quans, order)
      
      new Conjunction (resultingQuans,
                       resultingSubst(arithConj),
                       resultingSubst(negatedConjs),
                       order)
    }
  }
  
  /**
   * Remove all quantifiers from <code>quans</code> that talk about variables
   * not occurring in <code>occurringVars</code>. The result is the filtered
   * sequence of quantifiers and a <code>Substitution</code> for adjusting the
   * remaing variables
   */
  private def eliminateUnusedQuans(occurringVars : Set[VariableTerm],
                                   quans : Seq[Quantifier],
                                   order : TermOrder)
                                          : (Seq[Quantifier], Substitution) = {
    val resultingQuans = new ArrayBuffer[Quantifier]
    var resultingSubst : Substitution = new IdentitySubst(order)
    
    for ((q, i) <- quans.elements.zipWithIndex) {
      if (occurringVars contains VariableTerm(i))
        resultingQuans += q
      else
        resultingSubst = ComposeSubsts(resultingSubst,
                                       new VariableShiftSubst(i+1, -1, order),
                                       order)
    }

    (resultingQuans, resultingSubst)
  }
   
  val TRUE : Conjunction =
    new Conjunction (List(), ArithConj.TRUE, NegatedConjunctions.TRUE, TermOrder.EMPTY)
  
  val FALSE : Conjunction =
    new Conjunction (List(), ArithConj.FALSE, NegatedConjunctions.TRUE, TermOrder.EMPTY)
                                            
}

/**
 * Class for representing (possibly quantified) conjunctions of arithmetic
 * literal (equations, inequalities), predicate literals and negated
 * <code>Conjunction</code>s. <code>quans</code> describes how the lowest
 * <code>quans.size</code> variables are quantified in the conjunction 
 * (<code>quans(0)</code>) is responsible for <code>VariableTerm(0)</code> and
 * is the innermost quantifier, etc).
 */
class Conjunction private (val quans : Seq[Quantifier],
                           val arithConj : ArithConj,
                           val negatedConjs : NegatedConjunctions,
                           val order : TermOrder)
                  extends Formula with SortedWithOrder[Conjunction] {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(Conjunction.AC,
                   (arithConj isSortedBy order) &&
                   (negatedConjs isSortedBy order) &&
                   // if the conjunction of negated conjunctions would have to
                   // be false, the arithmetic conjunction is set to false
                   // instead
                   (!arithConj.isFalse || negatedConjs.isEmpty) &&
                   !negatedConjs.isFalse &&
                   // simple literals in the set of negated conjunctions can be
                   // normalised away
                   !negatedConjs.containsLiteral &&
                   // there must be no double-negations
                   !negatedConjs.containsNegatedConjunction &&
                   // there must not be any unused quantifiers
                   Logic.forall(0, quans.size,
                                (i) => variablesUnderQuans contains VariableTerm(i)) &&
                   // we do not want nested quantifiers through nested
                   // conjunctions; in this case, the quantifiers should be
                   // pulled out to the top-level
                   !(arithConj.isTrue && negatedConjs.size == 1 &&
                     !negatedConjs(0).quans.isEmpty))
  //////////////////////////////////////////////////////////////////////////////

  def sortBy(newOrder : TermOrder) : Conjunction = {
    if (isSortedBy(newOrder)) {
      this
    } else {
      // TODO: can the reordering possibly break the class invariants?
      Conjunction.createFromNormalised (quans,
                                        arithConj sortBy newOrder,
                                        negatedConjs sortBy newOrder,
                                        newOrder)
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
   * Perform propagation between the different parts of the conjunction. Right
   * now, this only delegates to the arithmetic conjunction (TODO: extend this)
   */
  def internallyPropagate : Conjunction =
    Conjunction(quans, arithConj.internallyPropagate, negatedConjs, order)

  //////////////////////////////////////////////////////////////////////////////

  private lazy val variablesUnderQuans : Set[VariableTerm] =
    arithConj.variables ++ negatedConjs.variables
    
  lazy val variables : Set[VariableTerm] =
    for (VariableTerm(i) <- variablesUnderQuans; if (i >= quans.size))
    yield VariableTerm(i - quans.size)

  lazy val constants : Set[ConstantTerm] =
    arithConj.constants ++ negatedConjs.constants

  //////////////////////////////////////////////////////////////////////////////

  def isTrue : Boolean = (arithConj.isTrue && negatedConjs.isTrue)

  def isFalse : Boolean = (arithConj.isFalse || negatedConjs.isFalse)

  /**
   * Return whether this conjunction actually only is a single literal
   * (a single, unquantified (in)equation, inequality or predicate literal)
   */
  def isLiteral : Boolean =
    (quans.isEmpty && arithConj.isLiteral && negatedConjs.isEmpty)

    
  /**
   * Return whether <code>this</code> is a divisibility judgement
   * <code>EX (n*_0 + t = 0)</code>
   */
  def isDivisibility : Boolean = {
    val res = (quans sameElements Array(Quantifier.EX)) &&
              (arithConj.positiveEqs.size == 1) &&
              arithConj.negativeEqs.isEmpty &&
              negatedConjs.isEmpty
              
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPost(Conjunction.AC,
         res == (!arithConj.positiveEqs.isEmpty &&
                 (arithConj.positiveEqs(0).variables contains VariableTerm(0)) && 
                 this == Conjunction.quantify(List(Quantifier.EX),
                                              EquationConj(arithConj.positiveEqs(0),
                                                           order),
                                              order)))
    ////////////////////////////////////////////////////////////////////////////
      
    res
  }

  /**
   * Return whether <code>this</code> is a negated divisibility judgement
   * <code>ALL (n*_0 + t != 0)</code>
   */
  def isNonDivisibility : Boolean = {
    val res = (quans sameElements Array(Quantifier.ALL)) &&
              arithConj.positiveEqs.isEmpty &&
              (arithConj.negativeEqs.size == 1) &&
              negatedConjs.isEmpty
                 
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPost(Conjunction.AC,
         res == (!arithConj.negativeEqs.isEmpty &&
                 (arithConj.negativeEqs(0).variables contains VariableTerm(0)) && 
                 this == Conjunction.quantify(List(Quantifier.ALL),
                                              NegEquationConj(arithConj.negativeEqs(0),
                                                              order),
                                              order)))
    ////////////////////////////////////////////////////////////////////////////
       
    res
  }

  def negate : Conjunction =
    Conjunction(List(), Iterator.single(NegatedConjunctions(this, order)), order)

  /**
   * Remove the <code>num</code> outermost quantifiers of this
   * <code>Conjunction</code>, without changing the matrix of the formula
   */
  def unquantify(num : Int) : Conjunction =
    Conjunction.createFromNormalised(quans.take(quans.size - num),
                                     arithConj, negatedConjs, order)
    
  /**
   * Update the arithmetic parts of this conjunction (without changing anything
   * else apart from the <code>TermOrder</code>) 
   */
  def updateArithConj(ac : ArithConj, newOrder : TermOrder) : Conjunction =
    Conjunction(quans, ac, negatedConjs, newOrder)

  /**
   * Update the positive equations of this conjunction (without changing anything
   * else apart from the <code>TermOrder</code>) 
   */
  def updatePositiveEqs(newEqs : EquationConj, newOrder : TermOrder) : Conjunction =
    Conjunction(quans,
                arithConj.updatePositiveEqs(newEqs, newOrder), negatedConjs,
                newOrder)

  /**
   * Return whether this conjunction actually is the negation of a single
   * conjunction
   */
  def isNegatedConjunction : Boolean =
    (quans.isEmpty && arithConj.isTrue && negatedConjs.size == 1)

  //////////////////////////////////////////////////////////////////////////////

  override def equals(that : Any) : Boolean = that match {
    case that : Conjunction => (this.quans sameElements that.quans) &&
                               this.arithConj == that.arithConj &&
                               this.negatedConjs == that.negatedConjs
    case _ => false
  }
  
  private lazy val hashCodeVal =
    Seqs.computeHashCode(this.quans, 982473, 3) +
    arithConj.hashCode + negatedConjs.hashCode

  override def hashCode = hashCodeVal
  
  override def toString : String = {
    if (isTrue) {
      "true"
    } else if (isFalse) {
      "false"
    } else {
      val quanPrefix = ("" /: quans)((p, q) => "" + q + " " + p)
      quanPrefix + "( " + ((arithConj.isTrue, negatedConjs.isTrue) match {
                           case (true, _) => "" + negatedConjs
                           case (_, true) => "" + arithConj
                           case _ => "" + arithConj + " & " + negatedConjs
                           }) + " )"
    }
  }

}
