package ap.terfor.equations;

import ap.basetypes.IdealInt
import ap.terfor.linearcombination.LinearCombination

/**
 * Abstract class for solving a conjunction of equations through column
 * operations.
 */
abstract class ColumnSolver(equations : EquationConj, order : TermOrder) {

  private var curEqs = equations
  private var curOrder = order
  
  /**
   * Decide whether the given linear combination describes an equation that can
   * be solved/simplified. This method has to be implemented by subclasses
   */
  protected def isSolvableEq(lc : LinearCombination) : Boolean
  
  /**
   * Introduce a new (small) symbol that is used to rename the leading term of
   * <code>lc</code> when performing a column operation. The result has to be
   * the small symbol and the (possibly extended) term ordering
   */
  protected def createSmallSym(lc : LinearCombination,
                               symDefinition : LinearCombination,
                               order : TermOrder) : (Term, TermOrder)

  private def chooseUnsolvedEq : Option[LinearCombination] =
    curEqs.find(lc => isSolvableEq(lc))

  private def defineSmallConst(lc : LinearCombination) : EquationConj = {
    val definition = lc.reduceWithLeadingCoeff
    val (smallConst, newOrder) = createSmallSym(lc, definition, curOrder)
    curOrder = newOrder
    val scLc = LinearCombination(smallConst, curOrder)
    val scDefLc = LinearCombination.sum(Array((IdealInt.ONE, definition),
                                              (IdealInt.MINUS_ONE, scLc)),
                                        curOrder)
    EquationConj(scDefLc, curOrder)
  }
  
   
  lazy val result = {
    var cont : Boolean = true    
    while (cont && !curEqs.isFalse) {
      chooseUnsolvedEq match {
      case Some(lc) => {
        val scDef = defineSmallConst(lc)
        curEqs = EquationConj.conj(Array(scDef, curEqs), curOrder)
      }
      case None => cont = false
      }
    }    

    if (curEqs.isFalse) curOrder = TermOrder.EMPTY

    (curEqs, curOrder)
  }
}
